﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class Chunk : MonoBehaviour {

    

    // rawData:
    public const int N_CHUNK_SIDELENGTH = 16; // FIXME: bad naming

    public const int N_CHUNK_SIDELENGTH_SQRD = 256; //16 * 16
    public const int N_CHUNK_BLOCKS=4096; //16*16*16

    public ChunkData chunk_data = null;

    protected MeshFilter meshFilter;
    
    protected List<Vector3> all_vertices;
    protected List<int> all_tris;
    protected List<Vector3> all_normals;
    protected List<Color> all_colors;

    protected List<Vector3> inner_vertices;
    protected List<int> inner_tris;
    protected List<Vector3> inner_normals;
    protected List<Color> inner_colors;

    // data for the border area:
    protected List<Vector3>[] border_vertices;
    protected List<int>[] border_tris;
    protected List<Vector3>[] border_normals;
    protected List<Color>[] border_colors;

    protected object border_handle = new object();

    public object borderHandle
    {
        get{
            return border_handle;
        }
    }


    public int[] ChunkPosition
    {
        get
        {
            return chunk_data.ChunkPosition;
        }
        set
        {
            chunk_data.ChunkPosition = value;
        }
    }

    public Vector3Int ChunkPositionVec
    {
        get
        {
            return chunk_data.ChunkPositionVec;
        }
        set
        {
            chunk_data.ChunkPositionVec = value;
        }
    }

    public int[] Data
    {
        get
        {
            return chunk_data.Data;
        }

    }

    public void DeleteBorderVertices(ChunkData.Direction dir)
    {
        int d = (int)dir;

        lock(border_handle)
        {
            border_vertices[d] = null;
            border_tris[d] = null;
            border_normals[d] = null;
            border_colors[d] = null;
        }
    }

    void Awake()
    {
        meshFilter = this.gameObject.AddComponent<MeshFilter>();

        meshFilter.mesh.MarkDynamic();

        all_vertices = new List<Vector3>();
        all_tris = new List<int>();
        all_normals = new List<Vector3>();
        all_colors = new List<Color>();

        inner_vertices = new List<Vector3>();
        inner_tris = new List<int>();
        inner_normals = new List<Vector3>();
        inner_colors = new List<Color>();

        border_vertices = new List<Vector3>[6];
        border_tris = new List<int>[6];
        border_normals = new List<Vector3>[6];
        border_colors = new List<Color>[6];

        for (int i = 0; i < 6; i++)
        {
            border_vertices[i] = null;
            border_tris[i] = null;
            border_normals[i] = null;
            border_colors[i] = null;
        }

    }

    // Use this for initialization
    void Start () {
        
    }


    public void Index2Coords(int i, int[] coords)
    {
        coords[0] = i % N_CHUNK_SIDELENGTH;
        coords[1] = (i / N_CHUNK_SIDELENGTH) % N_CHUNK_SIDELENGTH;
        coords[2] = i / (N_CHUNK_SIDELENGTH * N_CHUNK_SIDELENGTH);
    }

    public int Coords2Index(int[] coords)
    {
        return coords[0] + coords[1] * N_CHUNK_SIDELENGTH + coords[2] * N_CHUNK_SIDELENGTH_SQRD;
    }

    protected Color CalculateColor(int[] coords)
    {
        return Color.blue + (((float)coords[1] + chunk_data.ChunkPosition[1] * N_CHUNK_SIDELENGTH) / (float)WorldManager.Instance.WorldHeight) * (Color.green - Color.blue);
    }

    protected void AddPlane(Vector3 normal,
                            Vector3 pos,
                            Color color,
                            List<Vector3> verts,
                            List<int> tris,
                            List<Vector3> normals,
                            List<Color> colors)
    {

        /*
        plane looks like this:
            b
           /|
          / |
         /  |
normal	a   |
 <------|-+---------+pos (center of cube)
        |   c 
        |  / 
        | /
        |/
        d
        
         */

        Vector3 a,b,c,d;

        Vector3 normal_flipped = Vector3.one - normal;
        float half_normal;

        if (normal.x * normal.x > 0)
        {
            half_normal = .5f * normal.x;
            a = new Vector3(half_normal,  half_normal,  half_normal);
            b = new Vector3(half_normal,  half_normal, -half_normal);
            c = new Vector3(half_normal, -half_normal, -half_normal);
            d = new Vector3(half_normal, -half_normal,  half_normal);
        }
        else if (normal.y * normal.y > 0)
        {
            half_normal = .5f * normal.y;
            a = new Vector3( half_normal, half_normal,  half_normal);
            b = new Vector3(-half_normal, half_normal,  half_normal);
            c = new Vector3(-half_normal, half_normal, -half_normal);
            d = new Vector3( half_normal, half_normal, -half_normal);
        }
        else
        {
            half_normal = .5f * normal.z;
            a = new Vector3( half_normal,  half_normal,  half_normal);
            b = new Vector3( half_normal, -half_normal,  half_normal);
            c = new Vector3(-half_normal, -half_normal,  half_normal);
            d = new Vector3(-half_normal,  half_normal,  half_normal);
        }

        a += pos;
        b += pos;
        c += pos;
        d += pos;

        int n = verts.Count;

        verts.Add(a);
        verts.Add(half_normal > 0 ? d : b); // forcing clockwise built triangles 
        verts.Add(c);
        verts.Add(half_normal > 0 ? b : d);

        tris.Add(n + 0);
        tris.Add(n + 1);
        tris.Add(n + 2);

        tris.Add(n + 2);
        tris.Add(n + 3);
        tris.Add(n + 0);

        normals.Add(normal);
        normals.Add(normal);
        normals.Add(normal);
        normals.Add(normal);

        colors.Add(color);
        colors.Add(color);
        colors.Add(color);
        colors.Add(color);
    }

    // Build the chunk mesh from ground up
    public void BuildInnerMesh()
    {

        inner_vertices.Clear();
        inner_tris.Clear();
        inner_normals.Clear();
        inner_colors.Clear();

        int i = 0;
        int[] pos = new int[3];
        int[] pos_buffer = new int[3];

        Index2Coords(i, pos);

        // iterate through the raw data:
        foreach (int d in Data)
        {
            int k = 0;
            if (d == 0)
            {
                // since there is air here, the adjacent blocks are visible. So we will draw the adjacent sides
                
                // manually go through 6 neighbors:

                pos_buffer[0] = pos[0];
                pos_buffer[1] = pos[1];
                pos_buffer[2] = pos[2];
                
                // x axis
                if (pos[0] > 0)
                {
                    pos_buffer[0] -= 1;
                    k = Coords2Index(pos_buffer);
                    if (Data[k] != 0)
                    {
                        AddPlane(Vector3.right, 
                                 new Vector3(pos_buffer[0], pos_buffer[1], pos_buffer[2]),
                                 CalculateColor(pos_buffer),
                                 inner_vertices,
                                 inner_tris,
                                 inner_normals,
                                 inner_colors);
                    }
                    pos_buffer[0] += 1;

                }

                if (pos[0] + 1 < N_CHUNK_SIDELENGTH)
                {
                    pos_buffer[0] += 1;
                    k = Coords2Index(pos_buffer);
                    if (Data[k] != 0)
                    {
                        AddPlane(Vector3.left,
                                 new Vector3(pos_buffer[0], pos_buffer[1], pos_buffer[2]),
                                 CalculateColor(pos_buffer),
                                 inner_vertices,
                                 inner_tris,
                                 inner_normals,
                                 inner_colors);
                    }
                    pos_buffer[0] -= 1;
                }

                // y axis
                if (pos[1] > 0)
                {
                    pos_buffer[1] -= 1;
                    k = Coords2Index(pos_buffer);
                    if (Data[k] != 0)
                    {
                        AddPlane(Vector3.up, 
                                 new Vector3(pos_buffer[0], pos_buffer[1], pos_buffer[2]),
                                 CalculateColor(pos_buffer),
                                 inner_vertices,
                                 inner_tris,
                                 inner_normals,
                                 inner_colors);
                    }
                    pos_buffer[1] += 1;

                }

                if (pos[1] + 1 < N_CHUNK_SIDELENGTH)
                {
                    pos_buffer[1] += 1;
                    k = Coords2Index(pos_buffer);
                    if (Data[k] != 0)
                    {
                        AddPlane(Vector3.down,
                                 new Vector3(pos_buffer[0], pos_buffer[1], pos_buffer[2]),
                                 CalculateColor(pos_buffer),
                                 inner_vertices,
                                 inner_tris,
                                 inner_normals,
                                 inner_colors);
                    }
                    pos_buffer[1] -= 1;
                }

                // z axis
                if (pos[2] > 0)
                {
                    pos_buffer[2] -= 1;
                    k = Coords2Index(pos_buffer);
                    if (Data[k] != 0)
                    {
                        AddPlane(Vector3.forward,
                                 new Vector3(pos_buffer[0], pos_buffer[1], pos_buffer[2]),
                                 CalculateColor(pos_buffer),
                                 inner_vertices,
                                 inner_tris,
                                 inner_normals,
                                 inner_colors);
                    }
                    pos_buffer[2] += 1;

                }

                if (pos[2] + 1 < N_CHUNK_SIDELENGTH)
                {
                    pos_buffer[2] += 1;
                    k = Coords2Index(pos_buffer);
                    if (Data[k] != 0)
                    {
                        AddPlane(Vector3.back, 
                                 new Vector3(pos_buffer[0], pos_buffer[1], pos_buffer[2]),
                                 CalculateColor(pos_buffer),
                                 inner_vertices,
                                 inner_tris,
                                 inner_normals,
                                 inner_colors);
                    }
                    pos_buffer[2] -= 1;
                }
            }

            i += 1;

            Index2Coords(i, pos);
            
        }
    }

    public void BuildBorderGeometry(ChunkData.Direction dir)
    {
        if (chunk_data.Neighbors[(int)dir] == null)
        {
            Debug.LogWarning("tried to build geometry from non existing chunk: " + (chunk_data.ChunkPositionVec + ChunkData.DirectionNormalsInt[(int)dir]).ToString());
            return;
        }

        // depending on direction iterate throug different dimensions

        int d  = (int) dir;
        int op_d = (d + 3) % 6;

        int[] pos = new int[3];

        int neighbor_index;
        int index;

        lock(border_handle)
        {

            border_vertices[d] = new List<Vector3>();
            border_tris[d] = new List<int>();
            border_normals[d] = new List<Vector3>();
            border_colors[d] = new List<Color>();

            // X

            if (dir == ChunkData.Direction.x_neg)
            {
                for (int y = 0; y < N_CHUNK_SIDELENGTH; y++)
                {
                    for (int z = 0; z < N_CHUNK_SIDELENGTH; z++)
                    {
                        pos[0] = N_CHUNK_SIDELENGTH - 1;
                        pos[1] = y;
                        pos[2] = z;

                        neighbor_index = Coords2Index(pos);
                        pos[0] = 0;
                        index = Coords2Index(pos);
                        pos[0] = -1;

                        if (chunk_data.Neighbors[d].Data[neighbor_index] != 0 && Data[index] == 0)
                        {
                            AddPlane(ChunkData.DirectionNormals[op_d],
                                    new Vector3(pos[0], pos[1], pos[2]),
                                    CalculateColor(pos),
                                    border_vertices[d],
                                    border_tris[d],
                                    border_normals[d],
                                    border_colors[d]);
                        }
                    }
                }
            }

            else if (dir == ChunkData.Direction.x_pos)
            {
                for (int y = 0; y < N_CHUNK_SIDELENGTH; y++)
                {
                    for (int z = 0; z < N_CHUNK_SIDELENGTH; z++)
                    {
                        pos[0] = 0;
                        pos[1] = y;
                        pos[2] = z;

                        neighbor_index = Coords2Index(pos);
                        pos[0] = N_CHUNK_SIDELENGTH - 1;
                        index = Coords2Index(pos);
                        pos[0] = N_CHUNK_SIDELENGTH;

                        if (chunk_data.Neighbors[d].Data[neighbor_index] != 0 && Data[index] == 0)
                        {
                            AddPlane(ChunkData.DirectionNormals[op_d],
                                    new Vector3(pos[0], pos[1], pos[2]),
                                    CalculateColor(pos),
                                    border_vertices[d],
                                    border_tris[d],
                                    border_normals[d],
                                    border_colors[d]);
                        }
                    }
                }
            }

            // Y

            else if (dir == ChunkData.Direction.y_neg)
            {
                for (int x = 0; x < N_CHUNK_SIDELENGTH; x++)
                {
                    for (int z = 0; z < N_CHUNK_SIDELENGTH; z++)
                    {
                        pos[0] = x;
                        pos[1] = N_CHUNK_SIDELENGTH - 1;
                        pos[2] = z;

                        neighbor_index = Coords2Index(pos);
                        pos[1] = 0;
                        index = Coords2Index(pos);
                        pos[1] = -1;

                        if (chunk_data.Neighbors[d].Data[neighbor_index] != 0 && Data[index] == 0)
                        {
                            AddPlane(ChunkData.DirectionNormals[op_d],
                                    new Vector3(pos[0], pos[1], pos[2]),
                                   CalculateColor(pos),
                                    border_vertices[d],
                                    border_tris[d],
                                    border_normals[d],
                                    border_colors[d]);
                        }
                    }
                }
            }

            else if (dir == ChunkData.Direction.y_pos)
            {
                for (int x = 0; x < N_CHUNK_SIDELENGTH; x++)
                {
                    for (int z = 0; z < N_CHUNK_SIDELENGTH; z++)
                    {
                        pos[0] = x;
                        pos[1] = 0;
                        pos[2] = z;

                        neighbor_index = Coords2Index(pos);
                        pos[1] = N_CHUNK_SIDELENGTH - 1;
                        index = Coords2Index(pos);
                        pos[1] = N_CHUNK_SIDELENGTH;

                        if (chunk_data.Neighbors[d].Data[neighbor_index] != 0 && Data[index] == 0)
                        {
                            AddPlane(ChunkData.DirectionNormals[op_d],
                                    new Vector3(pos[0], pos[1], pos[2]),
                                    CalculateColor(pos),
                                    border_vertices[d],
                                    border_tris[d],
                                    border_normals[d],
                                    border_colors[d]);
                        }
                    }
                }
            }

            // Z

            else if (dir == ChunkData.Direction.z_neg)
            {
                for (int x = 0; x < N_CHUNK_SIDELENGTH; x++)
                {
                    for (int y = 0; y < N_CHUNK_SIDELENGTH; y++)
                    {
                        pos[0] = x;
                        pos[1] = y;
                        pos[2] = N_CHUNK_SIDELENGTH - 1;

                        neighbor_index = Coords2Index(pos);
                        pos[2] = 0;
                        index = Coords2Index(pos);
                        pos[2] = -1;
                        
                        if (chunk_data.Neighbors[d].Data[neighbor_index] != 0 && Data[index] == 0)
                        {
                            AddPlane(ChunkData.DirectionNormals[op_d],
                                    new Vector3(pos[0], pos[1], pos[2]),
                                    CalculateColor(pos),
                                    border_vertices[d],
                                    border_tris[d],
                                    border_normals[d],
                                    border_colors[d]);
                        }
                    }
                }
            }

            else if (dir == ChunkData.Direction.z_pos)
            {
                for (int x = 0; x < N_CHUNK_SIDELENGTH; x++)
                {
                    for (int y = 0; y < N_CHUNK_SIDELENGTH; y++)
                    {
                        pos[0] = x;
                        pos[1] = y;
                        pos[2] = 0;

                        neighbor_index = Coords2Index(pos);
                        pos[2] = N_CHUNK_SIDELENGTH - 1;
                        index = Coords2Index(pos);
                        pos[2] = N_CHUNK_SIDELENGTH;

                        if (chunk_data.Neighbors[d].Data[neighbor_index] != 0 && Data[index] == 0)
                        {
                            AddPlane(ChunkData.DirectionNormals[op_d],
                                    new Vector3(pos[0], pos[1], pos[2]),
                                    CalculateColor(pos),
                                    border_vertices[d],
                                    border_tris[d],
                                    border_normals[d],
                                    border_colors[d]);
                        }
                    }
                }
            }

        }

        
    }

    public void ConcatenateGeometry()
    {

        int n_inner = inner_vertices.Count;

        all_vertices = new List<Vector3>(inner_vertices);
        all_tris = new List<int>(inner_tris);
        all_normals = new List<Vector3>(inner_normals);
        all_colors = new List<Color>(inner_colors);

        int c_verts = n_inner;
        int c_tris = inner_tris.Count;

        lock (border_handle)
        {
            //get vertices count of border areas:
            for (int i = 0; i < 6; i++)
            {
                if (border_vertices[i] != null)
                {
                    all_vertices.AddRange(border_vertices[i]);
                    all_tris.AddRange(border_tris[i]);
                    all_normals.AddRange(border_normals[i]);
                    all_colors.AddRange(border_colors[i]);

                    // correct triangle indices
                    for (int tri_i = c_tris; tri_i < c_tris + border_tris[i].Count; tri_i++)
                    {
                        all_tris[tri_i] += c_verts;
                    }

                    c_tris += border_tris[i].Count;
                    c_verts += border_vertices[i].Count;
                }
            }

        }

    }

    public void AssignMesh()
    {
        meshFilter.mesh.Clear();
        meshFilter.mesh.vertices = all_vertices.ToArray();
        meshFilter.mesh.triangles = all_tris.ToArray();
        meshFilter.mesh.normals = all_normals.ToArray();
        meshFilter.mesh.colors = all_colors.ToArray();
    }

    void OnDestroy()
    {
        // Disconnect from all neighbors
        chunk_data.DisconnectChunkData(ChunkData.Direction.x_neg);
        chunk_data.DisconnectChunkData(ChunkData.Direction.y_neg);
        chunk_data.DisconnectChunkData(ChunkData.Direction.z_neg);
        chunk_data.DisconnectChunkData(ChunkData.Direction.x_pos);
        chunk_data.DisconnectChunkData(ChunkData.Direction.y_pos);
        chunk_data.DisconnectChunkData(ChunkData.Direction.z_pos);
        
    }
        // Update is called once per frame
    void Update () {
        //Pulse();
    }
}
