﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

public class ChunkData {

	public enum ChunkGeometryType
    {
		Unknown,
        Mixed,
        Full,
        Empty
    };

	public enum Direction {x_neg, y_neg, z_neg, x_pos, y_pos, z_pos};

    protected static Vector3Int[] direction_normals_int = new Vector3Int[] {Vector3Int.left,
                                            Vector3Int.down,
                                            new Vector3Int(0,0,-1),
                                            Vector3Int.right,
                                            Vector3Int.up,
                                            new Vector3Int(0,0,1)};

    protected static Vector3[] direction_normals = new Vector3[] {Vector3.left,
                                            Vector3.down,
                                            new Vector3(0,0,-1),
                                            Vector3.right,
                                            Vector3.up,
                                            new Vector3(0,0,1)};
    
    public static Vector3Int[] DirectionNormalsInt
    {
        get
        {
            return direction_normals_int;
        }
    }

    public static Vector3[] DirectionNormals
    {
        get
        {
            return direction_normals;
        }
    }

	public const int N_CHUNK_SIDELENGTH = 16; // FIXME: bad naming

    public const int N_CHUNK_SIDELENGTH_SQRD = 256; //16 * 16
    public const int N_CHUNK_BLOCKS=4096; //16*16*16

	private int[] data = new int[N_CHUNK_BLOCKS];

	private ChunkGeometryType m_chunk_geometry_type = ChunkGeometryType.Unknown;

	private ChunkGeometryType[] m_chunk_geometry_type_borders = new ChunkGeometryType[] {ChunkGeometryType.Unknown,
																						 ChunkGeometryType.Unknown,
																						 ChunkGeometryType.Unknown,
																						 ChunkGeometryType.Unknown,
																						 ChunkGeometryType.Unknown,
																						 ChunkGeometryType.Unknown};

	private bool m_is_built_necessary = false;


	private Chunk m_chunk = null;

	private object neighbor_lock = new object();

	private ChunkData[] m_neighbors = new ChunkData[] {null, null, null, null, null, null};

	protected int[] chunk_position =  new int[3] {0,0,0};
    protected Vector3Int chunk_position_vec;

	protected int m_connected_neighbors = 0;
	protected bool m_is_fully_connected = false;

    public int[] ChunkPosition
    {
        get
        {
            return chunk_position;
        }
        set
        {
            chunk_position = value;
            chunk_position_vec = new Vector3Int(value[0], value[1], value[2]);
        }
    }

    public Vector3Int ChunkPositionVec
    {
        get
        {
            return chunk_position_vec;
        }
        set
        {
            chunk_position_vec = value;
            chunk_position = new int[3] {value.x, value.y, value.z};
        }
    }

	public int ConnectedNeighbors
	{
		get
		{
			int val;
			lock(neighbor_lock)
			{
				val = m_connected_neighbors;
			}
			return val;
		}
	}

	public bool IsFullyConnected
	{
		get
		{
			bool val;
			lock(neighbor_lock)
			{
				val = m_is_fully_connected;
			}
			return val;
		}
	}

	public bool isBuiltNecessary
	{
		get
		{
			return m_is_built_necessary;
		}
	}

	public int[] Data
	{
		get{
			return data;
		}
	}

	public Chunk ParentChunk
	{
		get
		{
			return m_chunk;
		}
		set
		{
			m_chunk = value;
		}
	}

	public ChunkGeometryType GeometryType
	{
		get
		{
			return m_chunk_geometry_type;
		}
	}

	public ChunkGeometryType[] GeometryTypeBorders
	{
		get
		{
			return m_chunk_geometry_type_borders;
		}
	}

	public ChunkData[] Neighbors
	{
		get
		{
			return m_neighbors;
		}
	}

	public ChunkData()
	{

	}

	public void Index2Coords(int i, int[] coords)
    {
        coords[0] = i % N_CHUNK_SIDELENGTH;
        coords[1] = (i / N_CHUNK_SIDELENGTH) % N_CHUNK_SIDELENGTH;
        coords[2] = i / (N_CHUNK_SIDELENGTH * N_CHUNK_SIDELENGTH);
    }

    public int Coords2Index(int[] coords)
    {
        return coords[0] + coords[1] * N_CHUNK_SIDELENGTH + coords[2] * N_CHUNK_SIDELENGTH_SQRD;
    }

	public ChunkData(MessageObject mo)
	{
		SetDataFromMessageObject(mo);
	}

	public void analyseData ()
    {
        int i = 0;
        if (data[i] == 0)
        {
			m_chunk_geometry_type = ChunkGeometryType.Empty;
            // first block is air
            while(i < data.Length)
            {
                if (data[i] != 0)
                {
                    m_chunk_geometry_type = ChunkGeometryType.Mixed;
					break;
                }
                i++;
            }
            
        }
        else
        {
			m_chunk_geometry_type = ChunkGeometryType.Full;
            // first block is something
            while (i < data.Length)
            {
                if (data[i] == 0)
                {
                    m_chunk_geometry_type = ChunkGeometryType.Mixed;
					break;
                }
                i++;
            }
            
        }

		// check borders:

		int[] p = {0,0,0};
		
		m_chunk_geometry_type_borders[(int)Direction.z_neg] = data[Coords2Index(p)] == 0 ? ChunkGeometryType.Empty : ChunkGeometryType.Full;
		m_chunk_geometry_type_borders[(int)Direction.x_neg] = m_chunk_geometry_type_borders[(int)Direction.y_neg] = m_chunk_geometry_type_borders[(int)Direction.z_neg];

		p[0] = N_CHUNK_SIDELENGTH - 1;
		p[1] = N_CHUNK_SIDELENGTH - 1;
		p[2] = N_CHUNK_SIDELENGTH - 1;

		m_chunk_geometry_type_borders[(int)Direction.z_pos] = data[Coords2Index(p)] == 0 ? ChunkGeometryType.Empty : ChunkGeometryType.Full;
		m_chunk_geometry_type_borders[(int)Direction.x_pos] = m_chunk_geometry_type_borders[(int)Direction.y_pos] = m_chunk_geometry_type_borders[(int)Direction.z_pos];

		p[0] = 0;
		p[1] = 0;
		p[2] = 0;
		bool empty = data[Coords2Index(p)] == 0;

		for(p[0] = 0; p[0] < N_CHUNK_SIDELENGTH; p[0]++)
		{
			for(p[1] = 0; p[1] < N_CHUNK_SIDELENGTH; p[1]++)
			{
				if (data[Coords2Index(p)] != 0 && empty || data[Coords2Index(p)] == 0 && !empty)
				{
					m_chunk_geometry_type_borders[(int)Direction.z_neg] = ChunkGeometryType.Mixed;
					break;
				}
			}
			if (m_chunk_geometry_type_borders[(int)Direction.z_neg] == ChunkGeometryType.Mixed)
			{
				break;
			}
		}

		p[0] = 0;
		p[1] = 0;
		empty = data[Coords2Index(p)] == 0;
		for(p[0] = 0; p[0] < N_CHUNK_SIDELENGTH; p[0]++)
		{
			for(p[2] = 0; p[2] < N_CHUNK_SIDELENGTH; p[2]++)
			{
				if (data[Coords2Index(p)] != 0 && empty || data[Coords2Index(p)] == 0 && !empty)
				{
					m_chunk_geometry_type_borders[(int)Direction.y_neg] = ChunkGeometryType.Mixed;
					break;
				}
			}
			if (m_chunk_geometry_type_borders[(int)Direction.y_neg] == ChunkGeometryType.Mixed)
			{
				break;
			}
		}

		p[0] = 0;
		p[2] = 0;
		empty = data[Coords2Index(p)] == 0;
		for(p[1] = 0; p[1] < N_CHUNK_SIDELENGTH; p[1]++)
		{
			for(p[2] = 0; p[2] < N_CHUNK_SIDELENGTH; p[2]++)
			{
				if (data[Coords2Index(p)] != 0 && empty || data[Coords2Index(p)] == 0 && !empty)
				{
					m_chunk_geometry_type_borders[(int)Direction.x_neg] = ChunkGeometryType.Mixed;
					break;
				}
			}
			if (m_chunk_geometry_type_borders[(int)Direction.x_neg] == ChunkGeometryType.Mixed)
			{
				break;
			}
		}

		p[1] = 0;
		p[2] = N_CHUNK_SIDELENGTH - 1;
		empty = data[Coords2Index(p)] == 0;
		for(p[0] = 0; p[0] < N_CHUNK_SIDELENGTH; p[0]++)
		{
			for(p[1] = 0; p[1] < N_CHUNK_SIDELENGTH; p[1]++)
			{
				if (data[Coords2Index(p)] != 0 && empty || data[Coords2Index(p)] == 0 && !empty)
				{
					m_chunk_geometry_type_borders[(int)Direction.z_pos] = ChunkGeometryType.Mixed;
					break;
				}
			}
			if (m_chunk_geometry_type_borders[(int)Direction.z_pos] == ChunkGeometryType.Mixed)
			{
				break;
			}
		}

		p[0] = 0;
		p[1] = N_CHUNK_SIDELENGTH - 1;
		empty = data[Coords2Index(p)] == 0;
		for(p[0] = 0; p[0] < N_CHUNK_SIDELENGTH; p[0]++)
		{
			for(p[2] = 0; p[2] < N_CHUNK_SIDELENGTH; p[2]++)
			{
				if (data[Coords2Index(p)] != 0 && empty || data[Coords2Index(p)] == 0 && !empty)
				{
					m_chunk_geometry_type_borders[(int)Direction.y_pos] = ChunkGeometryType.Mixed;
					break;
				}
			}
			if (m_chunk_geometry_type_borders[(int)Direction.y_pos] == ChunkGeometryType.Mixed)
			{
				break;
			}
		}

		p[0] = N_CHUNK_SIDELENGTH - 1;
		p[2] = 0;
		empty = data[Coords2Index(p)] == 0;
		for(p[1] = 0; p[1] < N_CHUNK_SIDELENGTH; p[1]++)
		{
			for(p[2] = 0; p[2] < N_CHUNK_SIDELENGTH; p[2]++)
			{
				if (data[Coords2Index(p)] != 0 && empty || data[Coords2Index(p)] == 0 && !empty)
				{
					m_chunk_geometry_type_borders[(int)Direction.x_pos] = ChunkGeometryType.Mixed;
					break;
				}
			}
			if (m_chunk_geometry_type_borders[(int)Direction.x_pos] == ChunkGeometryType.Mixed)
			{
				break;
			}
		}
		m_is_built_necessary = m_chunk_geometry_type == ChunkGeometryType.Mixed;
		
    }

	public void SetDataFromMessageObject(MessageObject mo)
    {
        if (mo.data == null || mo.position == null)
        {
            throw new System.Exception("received empty data block");
        }

        if (mo.data.Length != 4096)
        {
            throw new System.Exception("bad data received");
        }

        data = mo.data;
		ChunkPositionVec = new Vector3Int(mo.position[2], mo.position[1], mo.position[0]);
    }

	public void ConnectChunkData(ChunkData cd, Direction dir)
	{
		int d = (int)dir;
		int op_d = (d+3)%6;
		lock(neighbor_lock)
		{
			if (m_neighbors[d] == null)
			{
				m_connected_neighbors++;
				m_is_fully_connected = (m_connected_neighbors == 6);
			}

			m_neighbors[d] = cd;
		}

		ChunkGeometryType borderType = cd.m_chunk_geometry_type_borders[op_d];

		if (borderType != ChunkGeometryType.Empty && m_chunk_geometry_type == ChunkGeometryType.Empty)
		{
			m_is_built_necessary = true;
		}
	}

	public void DisconnectChunkData(Direction dir)
	{
		int d = (int)dir;
		int op_d = (d+3)%6;

		lock (neighbor_lock)
		{
			if (m_neighbors[d] == null)
			{
				return;
			}

			if (m_chunk != null)
			{
				m_chunk.DeleteBorderVertices(dir);
			}
			if (m_neighbors[d].m_chunk != null)
			{
				m_neighbors[d].m_chunk.DeleteBorderVertices((Direction) op_d);
			}

			m_neighbors[d].m_neighbors[op_d] = null;
			m_neighbors[d] = null;

			m_connected_neighbors--;
			m_is_fully_connected = false;
		}
		
	}


}
